use std::{collections::HashMap, convert::Infallible, error::Error, rc::Rc};

use crate::{
    DeBrujinAst,
    parse::{Ast, Constant},
    types::{Ident, PrimitiveType, TaggedType, Type, TypeTag},
    vec_map::VecMap,
};

#[derive(Debug)]
pub enum InferError {
    NotAFunction,
    MismatchedType,
    NotInContext,
    ExpectedConreteType,
    ExpectedTypeWithTag,
}

#[cfg(test)]
mod test;

pub fn infer_type(
    gamma: &HashMap<Ident, TaggedType>,
    ast: DeBrujinAst,
) -> Result<TaggedType, InferError> {
    infer_type_debrujin_int(gamma, Rc::new(VecMap::new()), ast)
}

fn infer_type_debrujin_int(
    gamma_free: &HashMap<Ident, TaggedType>,
    mut gamma_bound: Rc<VecMap<usize, TaggedType>>,
    ast: DeBrujinAst,
) -> Result<TaggedType, InferError> {
    match ast {
        DeBrujinAst::Abstraction(_, arg_type, ast) => {
            let gamma_ref = Rc::make_mut(&mut gamma_bound);
            gamma_ref.map_keys(|i| *i += 1);
            gamma_ref.insert(1, arg_type.clone().into());

            let out_type = infer_type_debrujin_int(gamma_free, gamma_bound, *ast)?;
            // TODO: Fix this hack
            let typ = arg_type.make_arrow(
                out_type
                    .to_concrete()
                    .ok_or(InferError::ExpectedConreteType)?,
            );
            Ok(typ)
        }
        DeBrujinAst::Application(lhs, rhs) => {
            let left_type = infer_type_debrujin_int(gamma_free, gamma_bound.clone(), *lhs)?;
            let Some((in_type, out_type)) = left_type.arrow() else {
                return Err(InferError::NotAFunction);
            };
            let Some(right_type) =
                infer_type_debrujin_int(gamma_free, gamma_bound.clone(), *rhs)?.to_concrete()
            else {
                return Err(InferError::ExpectedConreteType);
            };
            let out_type = match in_type.matches(&right_type) {
                (true, None) => out_type,
                (true, Some(a)) => out_type.specialize(&a, &right_type),
                (false, _) => return Err(InferError::MismatchedType),
            };
            Ok(out_type)
        }
        DeBrujinAst::FreeVariable(x) => gamma_free.get(&x).cloned().ok_or(InferError::NotInContext),
        // compiler bug if not present
        DeBrujinAst::BoundVariable(n) => Ok(gamma_bound.get(&n).cloned().unwrap()),
        DeBrujinAst::Constant(Constant::Nat(_)) => Ok(Type::Primitive(PrimitiveType::Nat).into()),
        DeBrujinAst::Constant(Constant::Bool(_)) => Ok(Type::Primitive(PrimitiveType::Bool).into()),
        DeBrujinAst::Constant(_) => unreachable!(),
    }
}
