use std::{
    collections::HashMap,
    io::{Write, stdin, stdout},
    rc::Rc,
};

use stlc_type_inference::{
    Ast, Builtin, DeBrujinAst, builtin, infer_type, parse_ast_str, parse_type_str,
};

macro_rules! repl_err {
    ($err:expr) => {{
        println!($err);
        print!("> ");
        stdout().flush().unwrap();
        continue;
    }};
}

fn main() {
    let mut builtins: HashMap<String, Rc<dyn Builtin>> = HashMap::new();
    builtins.insert("add".to_string(), Rc::new(builtin::AddOp));
    builtins.insert("sub".to_string(), Rc::new(builtin::SubOp));
    builtins.insert("mul".to_string(), Rc::new(builtin::MulOp));
    builtins.insert("if".to_string(), Rc::new(builtin::OpCond));
    let mut gamma = HashMap::new();
    for (k, v) in &builtins {
        gamma.insert(k.clone(), v.r#type());
    }
    print!("> ");
    stdout().flush().unwrap();
    for line in stdin().lines() {
        let line = line.unwrap();
        if let Some(tl) = line.strip_prefix(':') {
            if let Some((cmd, expr)) = tl.split_once(' ') {
                match cmd {
                    "t" => match parse_ast_str(expr) {
                        Ok(a) => match infer_type(&gamma, a.into()) {
                            Ok(t) => println!("{t}"),
                            Err(e) => repl_err!("Could not infer type {e:?}"),
                        },
                        Err(e) => repl_err!("Parse error {e}"),
                    },
                    "ctx" => {
                        if let Some((ident, typ)) = expr.split_once(':') {
                            let ident: String =
                                ident.chars().filter(|c| !c.is_whitespace()).collect();
                            if !is_ident(&ident) {
                                repl_err!("{ident} is not a valid identifer");
                            }
                            let typ = match parse_type_str(&typ) {
                                Ok(t) => t,
                                Err(e) => repl_err!("type could not be parsed {e}"),
                            };

                            if !gamma.contains_key(&ident) {
                                println!("Added {ident} with type {typ} to the context");
                                gamma.insert(ident, typ.into());
                            } else {
                                println!("Cannot override existing ctx");
                            }
                        }
                    }
                    c => println!("Unknown command {c}"),
                }
            } else {
                match tl {
                    "t" => println!(":t need an expr as an argument"),
                    "ctx" => println!(":ctx needs an ident and a type as argument"),
                    c => println!("Unknown command {c}"),
                }
            }
        } else {
            let ast = match parse_ast_str(&line) {
                Ok(a) => a,
                Err(e) => repl_err!("Parse error {e}"),
            };

            let ast: DeBrujinAst = ast.into();
            let typ = match infer_type(&gamma, ast.clone()) {
                Ok(t) => t,
                Err(e) => repl_err!("Could not infer type {e:?}"),
            };
            let ast = ast.beta_reduce();
            let ast = ast.reduce_builtins(&builtins);
            let ast: Ast = ast.into();
            println!("{ast} : {typ}")
        }
        print!("> ");
        stdout().flush().unwrap();
    }
}

fn is_ident(s: &str) -> bool {
    s.starts_with(|c: char| c.is_alphabetic()) && s.chars().all(|c| c.is_alphanumeric())
}
