use std::{
    fmt::Display,
    num::{ParseFloatError, ParseIntError},
    str::ParseBoolError,
};

use crate::types::{Ident, PrimitiveType, TaggedType, Type, TypeTag};
use lalrpop_util::{ParseError as LALRPopError, lalrpop_mod};
use tokenize::Lexer;

#[cfg(test)]
mod test;

mod tokenize;

lalrpop_mod!(parser);

#[derive(Default, Debug, Clone, PartialEq)]
pub enum ParseError {
    UnknownTypeTag(String),
    InvalidInteger(ParseIntError),
    InvalidFloat(ParseFloatError),
    InvalidBool(ParseBoolError),
    #[default]
    InvalidToken,
}

impl Display for ParseError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{self:?}") //TODO better impl
    }
}

#[derive(Debug, Clone, PartialEq)]
pub enum Constant {
    Nat(usize),
    Float(f64),
    Bool(bool),
}

#[derive(Debug, Clone, PartialEq)]
pub enum Ast {
    Abstraction(Ident, TaggedType, Box<Ast>), // \0:1.2
    Application(Box<Ast>, Box<Ast>),          // 0 1
    Variable(Ident),                          // x
    Constant(Constant),                       // true | false | n
}

pub fn parse_ast_str(src: &str) -> Result<Ast, LALRPopError<usize, tokenize::Token, ParseError>> {
    let parser = parser::AstParser::new();
    parser.parse(Lexer::new(src))
}

pub fn parse_type_str(
    src: &str,
) -> Result<TaggedType, LALRPopError<usize, tokenize::Token, ParseError>> {
    let parser = parser::TaggedTypeParser::new();
    parser.parse(Lexer::new(src))
}

impl TryFrom<String> for TypeTag {
    type Error = ParseError;

    fn try_from(value: String) -> Result<Self, Self::Error> {
        match value.as_str() {
            "Num" => Ok(TypeTag::Num),
            "Any" => Ok(TypeTag::Any),
            _ => Err(ParseError::UnknownTypeTag(value)),
        }
    }
}

impl Type {
    pub fn from_name(name: String) -> Self {
        match name.as_str() {
            "Nat" => Self::Primitive(PrimitiveType::Nat),
            "Float" => Self::Primitive(PrimitiveType::Float),
            "Bool" => Self::Primitive(PrimitiveType::Bool),
            a => a.into(),
        }
    }
}

impl Display for Ast {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Ast::Abstraction(var, typ, ast) => write!(f, "(\\{var}:{typ}.{ast})"),
            Ast::Application(lhs, rhs) => write!(f, "{lhs} {rhs}"),
            Ast::Variable(v) => write!(f, "{v}"),
            Ast::Constant(constant) => write!(f, "{constant}"),
        }
    }
}

impl Display for Constant {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Constant::Nat(n) => write!(f, "{n}"),
            Constant::Bool(b) => write!(f, "{b}"),
            Constant::Float(fl) => write!(f, "{fl}"),
        }
    }
}
